"""Client and utilities to work with the Magento 2 REST API."""

from magento.attributes import (
    get_custom_attribute, get_boolean_custom_attribute, get_custom_attributes_dict,
    set_custom_attribute, set_custom_attributes, serialize_attribute_value,
    delete_custom_attribute, delete_custom_attributes,
    CATEGORY_ENTITY_TYPE_ID, CATALOG_PRODUCT_ENTITY_TYPE_ID,
)
from magento.client import Magento
from magento.dates import DATE_ISO_8601_FORMAT, format_datetime, parse_datetime
from magento.exceptions import MagentoException, MagentoAssertionError
from magento.order_helpers import is_order_on_hold, is_order_cash_on_delivery, get_order_shipping_address
from magento.queries import Query, make_field_value_query, make_search_query
from magento.types import (
    AttributeOption,
    MagentoEntity,
    BasePrice,
    Category,
    Customer,
    CustomAttributeDict,
    DeleteCouponsResponseDict,
    ErrorDict,
    MediaEntry,
    Order,
    PriceUpdateResultDict,
    Product,
    ShippingAddress,
    Sku,
    SourceItem,
    WithExtensionAttributesDict,
)
from magento.version import __version__

__all__ = [
    "AttributeOption",
    "BasePrice",
    "Category",
    "Customer",
    "CustomAttributeDict",
    "DATE_ISO_8601_FORMAT",
    "DISABLED_PRODUCT",
    "DeleteCouponsResponseDict",
    "ENABLED_PRODUCT",
    "ErrorDict",
    "IMAGE_MIME_TYPES",
    "Magento",
    "MagentoAssertionError",
    "MagentoEntity",
    "MagentoException",
    "MediaEntry",
    "Order",
    "PriceUpdateResultDict",
    "Product",
    "Query",
    "ROOT_CATEGORY_ID",
    "ShippingAddress",
    "Sku",
    "SourceItem",
    "VISIBILITY_IN_CATALOG",
    "VISIBILITY_IN_CATALOG_AND_SEARCH",
    "VISIBILITY_IN_SEARCH",
    "VISIBILITY_NOT_VISIBLE",
    "__version__",
    "delete_custom_attribute",
    "delete_custom_attributes",
    "format_datetime",
    "get_boolean_custom_attribute",
    "get_custom_attribute",
    "get_custom_attributes_dict",
    "get_order_shipping_address",
    "is_order_cash_on_delivery",
    "is_order_on_hold",
    "make_field_value_query",
    "make_search_query",
    "parse_datetime",
    "serialize_attribute_value",
    "set_custom_attribute",
    "set_custom_attributes",
    "WithExtensionAttributesDict",
]

ROOT_CATEGORY_ID = 1

# Magento's visibility options
# https://devdocs.magento.com/guides/v2.4/rest/tutorials/configurable-product/create-configurable-product.html
VISIBILITY_IN_CATALOG_AND_SEARCH = 4
VISIBILITY_IN_CATALOG = 2
VISIBILITY_IN_SEARCH = 3
VISIBILITY_NOT_VISIBLE = 1

# Magento product statuses
# https://magento.stackexchange.com/q/10693/92968
ENABLED_PRODUCT = 1
DISABLED_PRODUCT = 2

# https://magento.stackexchange.com/a/244151/92968
# https://github.com/magento/magento2/blob/ef6d9c80/lib/internal/Magento/Framework/Image/Adapter/Gd2.php#L30-L34
IMAGE_MIME_TYPES = {"image/gif", "image/jpeg", "image/png", "image/xbm", "image/wbmp"}



import sys
from typing import Union, TypedDict, Literal, List, Any, Dict

from api_session import JSONDict

if sys.version_info >= (3, 11):
    from typing import NotRequired
else:
    # noinspection PyUnreachableCode
    from typing_extensions import NotRequired

__all__ = (
    'AttributeOption',
    'BasePrice',
    'Category',
    'Customer',
    'CustomAttributeDict',
    'DeleteCouponsResponseDict',
    'MagentoEntity',
    'MediaEntry',
    'Order',
    'PathId',
    'ErrorDict',
    'PriceUpdateResultDict',
    'Product',
    'ShippingAddress',
    'Sku',
    'SourceItem',
    'SourceItemIn',
    'WithExtensionAttributesDict',
)

MagentoEntity = JSONDict

PathId = Union[int, str]
Sku = str

# TODO: proper types
AttributeOption = MagentoEntity
Category = MagentoEntity
Customer = MagentoEntity
MediaEntry = MagentoEntity
Order = MagentoEntity
ProductLink = MagentoEntity


class WithExtensionAttributesDict(TypedDict):
    """A dict with an extension_attributes key."""
    extension_attributes: Dict[str, Any]


class CustomAttributeDict(TypedDict):
    """A custom attribute dict, as found on products."""
    attribute_code: str
    value: Union[str, List[str], None]
    """A string value or a list of strings. This can be ``None`` to delete a custom attribute,
    but it will never be ``None`` in a Magento response.
    """


# Products
# ========

class Product(WithExtensionAttributesDict):
    """A product."""
    id: int
    sku: str
    name: str
    status: int
    attribute_set_id: int
    created_at: str
    updated_at: str
    custom_attributes: List[CustomAttributeDict]
    media_gallery_entries: List[MediaEntry]
    options: List[Any]
    price: float
    product_links: List[ProductLink]
    tier_prices: List[Any]
    type_id: str
    visibility: int
    weight: NotRequired[float]


# Source items
# ============
# See https://developer.adobe.com/commerce/webapi/rest/inventory/manage-source-items/

class SourceItemIn(TypedDict):
    """Input source item."""
    sku: Sku
    source_code: str
    quantity: NotRequired[int]
    status: NotRequired[int]


class SourceItem(TypedDict):
    """Source item as returned by Magento."""
    sku: Sku
    source_code: str
    quantity: int
    status: int


# Prices
# ======

class BasePrice(TypedDict):
    """Base price dict."""
    price: Union[int, float]
    store_id: int
    sku: Sku


# Orders
# ======

class ShippingAddress(TypedDict):
    """Shipping address dict."""
    address_type: Literal["shipping"]
    entity_id: int
    parent_id: int
    firstname: str
    lastname: str
    email: str
    street: List[str]
    city: str
    postcode: str
    region: NotRequired[str]
    region_code: str
    region_id: int
    country_id: str
    telephone: str
    company: NotRequired[str]


# Other types
# ===========

class DeleteCouponsResponseDict(TypedDict):
    """Response from the `coupons/deleteByIds` endpoint."""
    failed_items: List[Any]
    missing_items: List[Any]


class ErrorDict(TypedDict):
    """Error dict."""
    message: str
    parameters: List[str]


class PriceUpdateResultDict(ErrorDict, WithExtensionAttributesDict):
    """Response from the `products/base-prices` endpoint."""
    pass



from magento.types import Order, ShippingAddress

__all__ = (
    'is_order_on_hold',
    'is_order_cash_on_delivery',
    'get_order_shipping_address',
)


def is_order_on_hold(order: Order) -> bool:
    """Test if an order is on hold."""
    return order["status"] == "holded" or "hold_before_state" in order


def is_order_cash_on_delivery(order: Order) -> bool:
    """Test if an order is paid with 'cash-on-delivery'."""
    # From Magento\OfflinePayments\Model\Cashondelivery::PAYMENT_METHOD_CASHONDELIVERY_CODE
    # For some reason Mypy is too dumb to see this is a boolean
    b: bool = order["payment"]["method"] == 'cashondelivery'
    return b


def get_order_shipping_address(order: Order) -> ShippingAddress:
    """Return the first shipping address of an order.

    Note the returned dict is a reference, so if you modify it, it modifies the order.
    Make a copy if you want to modify the address without affecting the order.
    """
    address: ShippingAddress = order["extension_attributes"]["shipping_assignments"][0]["shipping"]["address"]
    return address



from datetime import datetime

DATE_ISO_8601_FORMAT = "%Y-%m-%d %H:%M:%S"


def format_datetime(dt: datetime) -> str:
    """Format a datetime for Magento."""
    # "2021-07-02 13:19:18.300700" -> "2021-07-02 13:19:18"
    return dt.isoformat(sep=" ").split(".", 1)[0]


def parse_datetime(s: str) -> datetime:
    """Parse a datetime string from Magento."""
    return datetime.strptime(s, DATE_ISO_8601_FORMAT)
